package main

import (
	"container/heap"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/antonholmquist/jason"
	"github.com/jasonlvhit/gocron"
	"io"
	"io/ioutil"
	"log"
	"os"
	"regexp"
	"strconv"
	"strings"
	"time"
)

var songs SongPriorityQueue

var credentials struct {
	APIURL   string
	UserName string
	Password string
}

func appendSong(wikiText string, author string, newSong string) string {
	const AUTHOR_MARK = "<!-- Lisääjä -->"
	const SONG_MARK = "<!-- Kappale -->"
	lines := strings.Split(wikiText, "\n")
	authorPrevIndex := -2

	changedLines := make([]string, 0, len(lines))
	for index, line := range lines {
		if strings.Index(line, AUTHOR_MARK) != -1 && strings.Index(line, author) != -1 {
			authorPrevIndex = index
		}
		if authorPrevIndex == (index-1) && strings.Index(line, SONG_MARK) != -1 {
			changedLines = append(changedLines, "| "+SONG_MARK+" "+newSong)
		} else {
			changedLines = append(changedLines, line)
		}
	}
	return strings.Join(changedLines, "\n")
}

func addSong(title string, week int, author string, song string) (bool, error) {
	wiki := CreateWikiClient(credentials.APIURL, credentials.UserName, credentials.Password)

	sections, err := wiki.GetWikiPageSections(title)
	if err != nil {
		return false, err
	}

	numberReg, _ := regexp.Compile("\\d+")
	for _, section := range sections {
		weekStr := numberReg.FindString(section.title)
		if weekStr != "" {
			weekNumber, _ := strconv.Atoi(weekStr)
			if weekNumber == week {
				wikiText, err := wiki.GetWikiPageSectionText(title, section.index)
				if err != nil {
					return false, err
				}
				changedWikiText := appendSong(wikiText, author, song)

				return wiki.EditWikiPageSection(title, section.index, changedWikiText,
					fmt.Sprintf("Added week %d song for %s", week, author))
			}
		}
	}
	return false, errors.New("Could not find matching section")
}

type SongsFile struct {
	Songs []*struct {
		Week   int
		Title  string
		Artist string
		URL    string
		Sync   bool
	}
}

func songSynced(syncedWeek int) error {
	var v SongsFile
	f, err := os.Open("songs.json")
	if err != nil {
		return err
	}
	defer f.Close()

	if err != nil {
		log.Fatal(err)
		return nil
	}
	dec := json.NewDecoder(f)
	dec.UseNumber()
	for {
		if err := dec.Decode(&v); err == io.EOF {
			break
		} else if err != nil {
			log.Fatal(err)
			return err
		}
	}

	synced := false
	for _, song := range v.Songs {
		if song.Week == syncedWeek {
			song.Sync = true
			synced = true
			jsonValue, err := json.Marshal(v)
			if err != nil {
				log.Fatal(err)
				return err
			}
			err = ioutil.WriteFile("songs.json", jsonValue, 0644)
			return err
		}
	}
	if !synced {
		return errors.New("No week matched from JSON for synced song")
	}
	return errors.New("Week not found")
}

func getSongs() (SongPriorityQueue, error) {
	f, err := os.Open("songs.json")
	if err != nil {
		return nil, err
	}
	defer f.Close()
	v, err := jason.NewObjectFromReader(f)
	if err != nil {
		return nil, err
	}
	songsArr, err := v.GetObjectArray("Songs")
	if err != nil {
		return nil, err
	}

	yearStart, _ := time.Parse(time.RFC3339, "2018-01-01T00:00:00+02:00")

	songs := make(SongPriorityQueue, len(songsArr))
	for index, songObj := range songsArr {
		title, err := songObj.GetString("Title")
		if err != nil {
			return nil, err
		}
		artist, err := songObj.GetString("Artist")
		if err != nil {
			return nil, err
		}
		url, err := songObj.GetString("URL")
		if err != nil {
			return nil, err
		}
		week64, err := songObj.GetInt64("Week")
		if err != nil {
			return nil, err
		}
		week := int(week64)
		sync, _ := songObj.GetBoolean("Sync")

		target := yearStart.AddDate(0, 0, (week-1)*7)
		songs[index] = &Song{
			time:  target,
			song:  "[" + url + " " + artist + " - " + title + "]",
			week:  week,
			sync:  sync,
			index: index,
		}
	}
	heap.Init(&songs)
	for len(songs) > 0 && songs[0].sync {
		heap.Pop(&songs)
	}
	return songs, nil
}

func task() {
	now := time.Now()
	if len(songs) > 0 && songs[0].time.Before(now) {
		fmt.Println("Time has passed for " + songs[0].song)
		success, err := addSong("Levyraati 2018", songs[0].week, "Lamperi", songs[0].song)
		if err != nil {
			log.Fatal(err)
		}
		if success {
			err := songSynced(songs[0].week)
			if err == nil {
				heap.Pop(&songs)
			} else {
				fmt.Println("Error received:", err)
			}
		}
	}
}

func initCreds() error {
	f, err := os.Open("credentials.json")
	if err != nil {
		return err
	}
	defer f.Close()

	if err != nil {
		log.Fatal(err)
		return err
	}
	dec := json.NewDecoder(f)
	for {
		if err := dec.Decode(&credentials); err == io.EOF {
			break
		} else if err != nil {
			log.Fatal(err)
		}
	}
	return nil
}

func main() {
	err := initCreds()
	if err != nil {
		panic(err)
	}
	songs, err = getSongs()
	if err != nil {
		panic(err)
	}
	gocron.Every(1).Second().Do(task)
	<-gocron.Start()
}
